@@ -1,6 +1,6 @@
 /* 
    HTTP Authentication routines
-   Copyright (C) 1999-2002, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2003, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -23,19 +23,36 @@
 /* HTTP Authentication, as per RFC2617.
  * 
  * TODO:
- *  - Improve cnonce? Make it really random using /dev/random or whatever?
  *  - Test auth-int support
  */
 
 #include "config.h"
 
+#include <sys/types.h>
+
+#ifdef HAVE_SYS_TIME_H
+#include <sys/time.h>
+#endif
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+#ifdef HAVE_UNISTD_H
+#include <unistd.h> /* for getpid() */
+#endif
+
+#ifdef WIN32
+#include <windows.h> /* for GetCurrentThreadId() etc */
+#endif
+
+#ifdef NEON_SSL
+#include <openssl/rand.h>
+#endif
 
 #include <time.h>
 
@@ -49,6 +66,15 @@
 #include "ne_uri.h"
 #include "ne_i18n.h"
 
+#ifdef HAVE_GSSAPI
+#ifdef HAVE_GSSAPI_GSSAPI_H /* MIT */
+#include <gssapi/gssapi.h>
+#include <gssapi/gssapi_generic.h>
+#else /* Heimdal. */
+#include <gssapi.h>
+#endif
+#endif
+
 /* TODO: should remove this eventually. Need it for
  * ne_pull_request_body. */
 #include "ne_private.h"
@@ -59,7 +85,8 @@
 /* The authentication scheme we are using */
 typedef enum {
     auth_scheme_basic,
-    auth_scheme_digest
+    auth_scheme_digest,
+    auth_scheme_gssapi
 } auth_scheme;
 
 typedef enum { 
@@ -79,7 +106,6 @@ typedef enum {
 struct auth_challenge {
     auth_scheme scheme;
     const char *realm;
-    const char *domain;
     const char *nonce;
     const char *opaque;
     unsigned int stale:1; /* if stale=true */
@@ -90,17 +116,6 @@ struct auth_challenge {
     struct auth_challenge *next;
 };
 
-static const char *qop_values[] = {
-    NULL,
-    "auth",
-    "auth-int"
-};
-static const char *algorithm_names[] = {
-    "MD5",
-    "MD5-sess",
-    NULL
-};
-
 static const struct auth_class {
     const char *id, *req_hdr, *resp_hdr, *resp_info_hdr, *fail_msg;
     int status_code, fail_code;
@@ -117,6 +132,13 @@ static const struct auth_class {
 /* Authentication session state. */
 typedef struct {
     ne_session *sess;
+
+    /* Which context will auth challenges be accepted? */
+    enum {
+        AUTH_ANY, /* ignore nothing. */
+        AUTH_CONNECT, /* only in response to a CONNECT request. */
+        AUTH_NOTCONNECT /* only in non-CONNECT responsees */
+    } context;
     
     /* Specifics for server/proxy auth. FIXME: need a better field
      * name! */
@@ -125,8 +147,8 @@ typedef struct {
     /* The scheme used for this authentication session */
     auth_scheme scheme;
     /* The callback used to request new username+password */
-    ne_request_auth reqcreds;
-    void *reqcreds_udata;
+    ne_auth_creds creds;
+    void *userdata;
 
     /*** Session details ***/
 
@@ -136,17 +158,18 @@ typedef struct {
     unsigned int can_handle:1;
     /* This used for Basic auth */
     char *basic; 
+#ifdef HAVE_GSSAPI
+    /* This used for GSSAPI auth */
+    char *gssapi_token;
+#endif
     /* These all used for Digest auth */
     char *realm;
     char *nonce;
     char *cnonce;
     char *opaque;
-    /* A list of domain strings */
-    unsigned int domain_count;
-    char **domain;
     auth_qop qop;
     auth_algorithm alg;
-    int nonce_count;
+    unsigned int nonce_count;
     /* The ASCII representation of the session's H(A1) value */
     char h_a1[33];
 
@@ -188,83 +211,72 @@ static void clean_session(auth_session *sess)
     NE_FREE(sess->cnonce);
     NE_FREE(sess->opaque);
     NE_FREE(sess->realm);
-    if (sess->domain_count > 0) {
-	split_string_free(sess->domain);
-	sess->domain_count = 0;
-    }
+#ifdef HAVE_GSSAPI
+    NE_FREE(sess->gssapi_token);
+#endif
 }
 
-/* Returns cnonce-value. We just use base64(time).
- * TODO: Could improve this? */
+/* Returns client nonce string. */
 static char *get_cnonce(void) 
 {
-    char *ret, *tmp;
-    tmp = ne_rfc1123_date(time(NULL));
-    ret = ne_base64(tmp, strlen(tmp));
-    free(tmp);
-    return ret;
+    char ret[33];
+    unsigned char data[256], tmp[16];
+    struct ne_md5_ctx hash;
+
+    ne_md5_init_ctx(&hash);
+
+#ifdef NEON_SSL
+    if (RAND_pseudo_bytes(data, sizeof data) >= 0)
+	ne_md5_process_bytes(data, sizeof data, &hash);
+    else {
+#endif
+    /* Fallback sources of random data: all bad, but no good sources
+     * are available. */
+
+    /* Uninitialized stack data; yes, happy valgrinders, this is
+     * supposed to be here. */
+    ne_md5_process_bytes(data, sizeof data, &hash);
+    
+#ifdef HAVE_GETTIMEOFDAY
+    {
+	struct timeval tv;
+	if (gettimeofday(&tv, NULL) == 0)
+	    ne_md5_process_bytes(&tv, sizeof tv, &hash);
+    }
+#else /* HAVE_GETTIMEOFDAY */
+    {
+	time_t t = time(NULL);
+	ne_md5_process_bytes(&t, sizeof t, &hash);
+    }
+#endif
+    {
+#ifdef WIN32
+	DWORD pid = GetCurrentThreadId();
+#else
+	pid_t pid = getpid();
+#endif
+	ne_md5_process_bytes(&pid, sizeof pid, &hash);
+    }
+
+#ifdef NEON_SSL
+    }
+#endif
+    
+    ne_md5_finish_ctx(&hash, tmp);
+    ne_md5_to_ascii(tmp, ret);
+
+    return ne_strdup(ret);
 }
 
 static int get_credentials(auth_session *sess, char *pwbuf) 
 {
-    return (*sess->reqcreds)(sess->reqcreds_udata, sess->realm, sess->attempt++,
-			    sess->username, pwbuf);
-}
-
-static int parse_domain(auth_session *sess, const char *domain) {
-    char **doms;
-    int count, ret;
-
-    doms = split_string_c(domain, ' ', NULL, " \r\n\t", &count);
-    if (doms != NULL) {
-	if (count > 0) {
-	    sess->domain = doms;
-	    sess->domain_count = count;
-	    ret = 0;
-	} else {
-	    free(doms);
-	    ret = -1;
-	}
-    } else {
-	ret = -1;
-    }
-
-    return ret;
-}
-
-/* Returns:
- *    0: if uri is in NOT in domain of session
- * else: uri IS in domain of session (or no domain known)
- */
-static int is_in_domain(auth_session *sess, const char *uri)
-{
-#if 1
-    return 1;
-#else
-    /* TODO: Need proper URI comparison for this to work. */
-    int ret, dom;
-    const char *abs_path;
-    if (sess->domain_count == 0) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "No domain, presuming okay.\n");
-	return 1;
-    }
-    ret = 0;
-    abs_path = uri_abspath(uri);
-    for (dom = 0; dom < sess->domain_count; dom++) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "Checking domain: %s\n", sess->domain[dom]);
-	if (uri_childof(sess->domain[dom], abs_path)) {
-	    ret = 1;
-	    break;
-	}
-    }
-    return ret;
-#endif
+    return sess->creds(sess->userdata, sess->realm, sess->attempt++,
+		       sess->username, pwbuf);
 }
 
 /* Examine a Basic auth challenge.
  * Returns 0 if an valid challenge, else non-zero. */
-static int 
-basic_challenge(auth_session *sess, struct auth_challenge *parms) 
+static int basic_challenge(auth_session *sess, struct auth_challenge *parms) 
 {
     char *tmp, password[NE_ABUFSIZ];
 
@@ -274,7 +286,7 @@ basic_challenge(auth_session *sess, struct auth_challenge *parms)
     }
 
     NE_DEBUG(NE_DBG_HTTPAUTH, "Got Basic challenge with realm [%s]\n", 
-	   parms->realm);
+	     parms->realm);
 
     clean_session(sess);
     
@@ -289,7 +301,7 @@ basic_challenge(auth_session *sess, struct auth_challenge *parms)
 
     tmp = ne_concat(sess->username, ":", password, NULL);
     sess->basic = ne_base64(tmp, strlen(tmp));
-    free(tmp);
+    ne_free(tmp);
 
     /* Paranoia. */
     memset(password, 0, sizeof password);
@@ -303,10 +315,76 @@ static char *request_basic(auth_session *sess)
     return ne_concat("Basic ", sess->basic, "\r\n", NULL);
 }
 
+#ifdef HAVE_GSSAPI
+/* Add GSSAPI authentication credentials to a request */
+static char *request_gssapi(auth_session *sess) 
+{
+    return ne_concat("GSS-Negotiate ", sess->gssapi_token, "\r\n", NULL);
+}
+
+static int get_gss_name(gss_name_t *server, auth_session *sess)
+{
+    unsigned int major_status, minor_status;
+    gss_buffer_desc token = GSS_C_EMPTY_BUFFER;
+
+    token.value = ne_concat("khttp@", sess->sess->server.hostname, NULL);
+    token.length = strlen(token.value);
+
+    major_status = gss_import_name(&minor_status, &token,
+                                   GSS_C_NT_HOSTBASED_SERVICE,
+                                   server);
+    return GSS_ERROR(major_status) ? -1 : 0;
+}
+
+/* Examine a GSSAPI auth challenge; returns 0 if a valid challenge,
+ * else non-zero. */
+static int 
+gssapi_challenge(auth_session *sess, struct auth_challenge *parms) 
+{
+    gss_ctx_id_t context;
+    gss_name_t server_name;
+    unsigned int major_status, minor_status;
+    gss_buffer_desc input_token = GSS_C_EMPTY_BUFFER;
+    gss_buffer_desc output_token = GSS_C_EMPTY_BUFFER;
+
+    clean_session(sess);
+
+    if (get_gss_name(&server_name, sess))
+        return -1;
+
+    major_status = gss_init_sec_context(&minor_status,
+                                        GSS_C_NO_CREDENTIAL,
+                                        &context,
+                                        server_name,
+                                        GSS_C_NO_OID,
+                                        GSS_C_DELEG_FLAG,
+                                        0,
+                                        GSS_C_NO_CHANNEL_BINDINGS,
+                                        &input_token,
+                                        NULL,
+                                        &output_token,
+                                        NULL,
+                                        NULL);
+    if (GSS_ERROR(major_status)) {
+        NE_DEBUG(NE_DBG_HTTPAUTH, "gss_init_sec_context failed.\n");
+        return -1;
+    }
+    
+    if (output_token.length == 0)
+        return -1;
+
+    sess->gssapi_token = ne_base64(output_token.value, output_token.length);
+
+    NE_DEBUG(NE_DBG_HTTPAUTH, 
+             "Base64 encoded GSSAPI challenge: %s.\n", sess->gssapi_token);
+    sess->scheme = auth_scheme_gssapi;
+    return 0;
+}
+#endif
+
 /* Examine a digest challenge: return 0 if it is a valid Digest challenge,
  * else non-zero. */
-static int digest_challenge(auth_session *sess,
-			    struct auth_challenge *parms) 
+static int digest_challenge(auth_session *sess, struct auth_challenge *parms) 
 {
     struct ne_md5_ctx tmp;
     unsigned char tmp_md5[16];
@@ -342,14 +420,7 @@ static int digest_challenge(auth_session *sess,
     sess->scheme = auth_scheme_digest;
     sess->nonce = ne_strdup(parms->nonce);
     sess->cnonce = get_cnonce();
-    if (parms->domain) {
-	if (parse_domain(sess, parms->domain)) {
-	    /* TODO: Handle the error? */
-	}
-    } else {
-	sess->domain = NULL;
-	sess->domain_count = 0;
-    }
+    /* TODO: add domain handling. */
     if (parms->opaque != NULL) {
 	sess->opaque = ne_strdup(parms->opaque); /* don't strip the quotes */
     }
@@ -433,10 +504,11 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
     if (sess->qop != auth_qop_none) {
 	sess->nonce_count++;
 	ne_snprintf(nc_value, 9, "%08x", sess->nonce_count);
-	NE_DEBUG(NE_DBG_HTTPAUTH, "Nonce count is %d, nc is [%s]\n", 
-	       sess->nonce_count, nc_value);
+	NE_DEBUG(NE_DBG_HTTPAUTH, "Nonce count is %u, nc is [%s]\n", 
+		 sess->nonce_count, nc_value);
     }
-    qop_value = qop_values[sess->qop];
+
+    qop_value = sess->qop == auth_qop_auth_int ? "auth-int" : "auth";
 
     /* Calculate H(A2). */
     ne_md5_init_ctx(&a2);
@@ -488,7 +560,7 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
 	 *    nc-value ":" unq(cnonce-value) ":" unq(qop-value) ":"
 	 */
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Have qop directive, digesting: [%s:%s:%s]\n",
-	       nc_value, sess->cnonce, qop_value);
+		 nc_value, sess->cnonce, qop_value);
 	ne_md5_process_bytes(nc_value, 8, &rdig);
 	ne_md5_process_bytes(":", 1, &rdig);
 	ne_md5_process_bytes(sess->cnonce, strlen(sess->cnonce), &rdig);
@@ -511,12 +583,13 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
     ret = ne_buffer_create();
 
     ne_buffer_concat(ret, 
-		   "Digest username=\"", sess->username, "\", "
-		   "realm=\"", sess->realm, "\", "
-		   "nonce=\"", sess->nonce, "\", "
-		   "uri=\"", req->uri, "\", "
-		   "response=\"", rdig_md5_ascii, "\", "
-		   "algorithm=\"", algorithm_names[sess->alg], "\"", NULL);
+		     "Digest username=\"", sess->username, "\", "
+		     "realm=\"", sess->realm, "\", "
+		     "nonce=\"", sess->nonce, "\", "
+		     "uri=\"", req->uri, "\", "
+		     "response=\"", rdig_md5_ascii, "\", "
+		     "algorithm=\"", sess->alg == auth_alg_md5 ? "MD5" : "MD5-sess", "\"", 
+		     NULL);
     
     if (sess->opaque != NULL) {
 	ne_buffer_concat(ret, ", opaque=\"", sess->opaque, "\"", NULL);
@@ -524,10 +597,9 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
 
     if (sess->qop != auth_qop_none) {
 	/* Add in cnonce and nc-value fields */
-	ne_buffer_concat(ret, 
-		       ", cnonce=\"", sess->cnonce, "\", "
-		       "nc=", nc_value, ", "
-		       "qop=\"", qop_values[sess->qop], "\"", NULL);
+	ne_buffer_concat(ret, ", cnonce=\"", sess->cnonce, "\", "
+			 "nc=", nc_value, ", "
+			 "qop=\"", qop_value, "\"", NULL);
     }
 
     ne_buffer_zappend(ret, "\r\n");
@@ -541,8 +613,7 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
  * then also return a leading space-separated token, as *value == NULL.
  * Otherwise, if return value is 0, *key and *value will be non-NULL.
  * If return value is non-zero, parsing has ended. */
-static int tokenize(char **hdr, char **key, char **value, 
-		    int ischall)
+static int tokenize(char **hdr, char **key, char **value, int ischall)
 {
     char *pnt = *hdr;
     enum { BEFORE_EQ, AFTER_EQ, AFTER_EQ_QUOTED } state = BEFORE_EQ;
@@ -587,6 +658,10 @@ static int tokenize(char **hdr, char **key, char **value,
 	}
     } while (*++pnt != '\0');
     
+    if (state == BEFORE_EQ && ischall && *key != NULL) {
+	*value = NULL;
+    }
+
     *hdr = pnt;
 
     /* End of string: */
@@ -609,7 +684,8 @@ static int verify_response(struct auth_request *req, auth_session *sess,
 	*cnonce = NULL, /* for the cnonce= value */
 	*nc = NULL, /* for the nc= value */
 	*qop_value = NULL;
-    int nonce_count, okay;
+    unsigned int nonce_count;
+    int okay;
     
     if (!req->will_handle) {
 	/* Ignore it */
@@ -617,7 +693,8 @@ static int verify_response(struct auth_request *req, auth_session *sess,
     }
     
     if (sess->scheme != auth_scheme_digest) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "Found Auth-Info header not in response to Digest credentials - dodgy.\n");
+	NE_DEBUG(NE_DBG_HTTPAUTH, "Found Auth-Info header not in response "
+		 " to Digest credentials - dodgy.\n");
 	return -1;
     }
 
@@ -645,10 +722,10 @@ static int verify_response(struct auth_request *req, auth_session *sess,
 	    cnonce = val;
 	} else if (strcasecmp(key, "nc") == 0) { 
 	    nc = val;
-	    if (sscanf(val, "%d", &nonce_count) != 1) {
+	    if (sscanf(val, "%x", &nonce_count) != 1) {
 		NE_DEBUG(NE_DBG_HTTPAUTH, "Couldn't find nonce count.\n");
 	    } else {
-		NE_DEBUG(NE_DBG_HTTPAUTH, "Got nonce_count: %d\n", nonce_count);
+		NE_DEBUG(NE_DBG_HTTPAUTH, "Got nonce_count: %u\n", nonce_count);
 	    }
 	}
     }
@@ -701,9 +778,9 @@ static int verify_response(struct auth_request *req, auth_session *sess,
 		if (qop != auth_qop_none) {
 		    /* Add in qop-value */
 		    NE_DEBUG(NE_DBG_HTTPAUTH, "Digesting qop-value [%s:].\n", 
-			   qop_value);
+			     qop_value);
 		    ne_md5_process_bytes(qop_value, strlen(qop_value), 
-				       &sess->stored_rdig);
+					 &sess->stored_rdig);
 		    ne_md5_process_bytes(":", 1, &sess->stored_rdig);
 		}
 		/* Digest ":" H(A2) */
@@ -712,10 +789,10 @@ static int verify_response(struct auth_request *req, auth_session *sess,
 		ne_md5_finish_ctx(&sess->stored_rdig, rdig_md5);
 		ne_md5_to_ascii(rdig_md5, rdig_md5_ascii);
 
-		NE_DEBUG(NE_DBG_HTTPAUTH, "Calculated response-digest of: [%s]\n",
-		       rdig_md5_ascii);
-		NE_DEBUG(NE_DBG_HTTPAUTH, "Given response-digest of:      [%s]\n",
-		       rspauth);
+		NE_DEBUG(NE_DBG_HTTPAUTH, "Calculated response-digest of: "
+			 "[%s]\n", rdig_md5_ascii);
+		NE_DEBUG(NE_DBG_HTTPAUTH, "Given response-digest of:      "
+			 "[%s]\n", rspauth);
 
 		/* And... do they match? */
 		okay = (strcasecmp(rdig_md5_ascii, rspauth) == 0)?0:-1;
@@ -731,11 +808,11 @@ static int verify_response(struct auth_request *req, auth_session *sess,
     if (nextnonce != NULL) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Found nextnonce of [%s].\n", nextnonce);
 	if (sess->nonce != NULL)
-	    free(sess->nonce);
+	    ne_free(sess->nonce);
 	sess->nonce = ne_strdup(nextnonce);
     }
 
-    free(hdr);
+    ne_free(hdr);
 
     return okay;
 }
@@ -774,9 +851,14 @@ static int auth_challenge(auth_session *sess, const char *value)
 	    } else if (strcasecmp(key, "digest") == 0) {
 		NE_DEBUG(NE_DBG_HTTPAUTH, "Digest scheme.\n");
 		chall->scheme = auth_scheme_digest;
+#ifdef HAVE_GSSAPI		
+	    } else if (strcasecmp(key, "gss-negotiate") == 0) {
+		NE_DEBUG(NE_DBG_HTTPAUTH, "GSSAPI scheme.\n");
+		chall->scheme = auth_scheme_gssapi;
+#endif
 	    } else {
 		NE_DEBUG(NE_DBG_HTTPAUTH, "Unknown scheme.\n");
-		free(chall);
+		ne_free(chall);
 		challenges = NULL;
 		break;
 	    }
@@ -799,8 +881,6 @@ static int auth_challenge(auth_session *sess, const char *value)
 	    chall->nonce = val;
 	} else if (strcasecmp(key, "opaque") == 0) {
 	    chall->opaque = val;
-	} else if (strcasecmp(key, "domain") == 0) {
-	    chall->domain = val;
 	} else if (strcasecmp(key, "stale") == 0) {
 	    /* Truth value */
 	    chall->stale = (strcasecmp(val, "true") == 0);
@@ -812,18 +892,17 @@ static int auth_challenge(auth_session *sess, const char *value)
 	    } else {
 		chall->alg = auth_alg_unknown;
 	    }
-	} else if (strcasecmp(val, "qop") == 0) {
-	    char **qops;
-	    int qop;
-	    qops = split_string(val, ',', NULL, " \r\n\t");
-	    chall->got_qop = 1;
-	    for (qop = 0; qops[qop] != NULL; qop++) {
-		if (strcasecmp(qops[qop], "auth") == 0) {
-		    chall->qop_auth = 1;
-		} else if (strcasecmp(qops[qop], "auth-int") == 0) {
-		    chall->qop_auth_int = 1;
-		}
-	    }
+	} else if (strcasecmp(key, "qop") == 0) {
+            /* iterate over each token in the value */
+            do {
+                const char *tok = ne_shave(ne_token(&val, ','), " \t");
+                
+                if (strcasecmp(tok, "auth") == 0) {
+                    chall->qop_auth = 1;
+                } else if (strcasecmp(tok, "auth-int") == 0 ) {
+                    chall->qop_auth_int = 1;
+                }
+            } while (val);
 	}
     }
 
@@ -831,26 +910,41 @@ static int auth_challenge(auth_session *sess, const char *value)
 
     /* Did we find any challenges */
     if (challenges == NULL) {
-	free(hdr);
+	ne_free(hdr);
 	return -1;
     }
     
     success = 0;
 
-    NE_DEBUG(NE_DBG_HTTPAUTH, "Looking for Digest challenges.\n");
-
-    /* Try a digest challenge */
+#ifdef HAVE_GSSAPI
+    NE_DEBUG(NE_DBG_HTTPAUTH, "Looking for GSSAPI.\n");
+    /* Try a GSSAPI challenge */
     for (chall = challenges; chall != NULL; chall = chall->next) {
-	if (chall->scheme == auth_scheme_digest) {
-	    if (!digest_challenge(sess, chall)) {
+	if (chall->scheme == auth_scheme_gssapi) {
+	    if (!gssapi_challenge(sess, chall)) {
 		success = 1;
 		break;
 	    }
 	}
     }
+#endif
+
+    /* Try a digest challenge */
+    if (!success) {
+        NE_DEBUG(NE_DBG_HTTPAUTH, "Looking for Digest challenges.\n");
+        for (chall = challenges; chall != NULL; chall = chall->next) {
+        	if (chall->scheme == auth_scheme_digest) {
+        	    if (!digest_challenge(sess, chall)) {
+        		success = 1;
+        		break;
+        	    }
+        	}
+        }
+    }
 
     if (!success) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "No good Digest challenges, looking for Basic.\n");
+	NE_DEBUG(NE_DBG_HTTPAUTH, 
+		 "No good Digest challenges, looking for Basic.\n");
 	for (chall = challenges; chall != NULL; chall = chall->next) {
 	    if (chall->scheme == auth_scheme_basic) {
 		if (!basic_challenge(sess, chall)) {
@@ -872,11 +966,11 @@ static int auth_challenge(auth_session *sess, const char *value)
 
     while (challenges != NULL) {
 	chall = challenges->next;
-	free(challenges);
+	ne_free(challenges);
 	challenges = chall;
     }
 
-    free(hdr);
+    ne_free(hdr);
 
     return !success;
 }
@@ -894,25 +988,31 @@ static void ah_create(ne_request *req, void *session, const char *method,
 		      const char *uri)
 {
     auth_session *sess = session;
-    struct auth_request *areq = ne_calloc(sizeof *areq);
-
-    NE_DEBUG(NE_DBG_HTTPAUTH, "ah_create, for %s\n", sess->spec->resp_hdr);
-
-    areq->method = method;
-    areq->uri = uri;
-    areq->request = req;
-
-    ne_add_response_header_handler(req, sess->spec->resp_hdr,
-				   ne_duplicate_header, &areq->auth_hdr);
+    int is_connect = strcmp(method, "CONNECT") == 0;
 
+    if (sess->context == AUTH_ANY ||
+        (is_connect && sess->context == AUTH_CONNECT) ||
+        (!is_connect && sess->context == AUTH_NOTCONNECT)) {
+        struct auth_request *areq = ne_calloc(sizeof *areq);
+        
+        NE_DEBUG(NE_DBG_HTTPAUTH, "ah_create, for %s\n", sess->spec->resp_hdr);
+        
+        areq->method = method;
+        areq->uri = uri;
+        areq->request = req;
+        
+        ne_add_response_header_handler(req, sess->spec->resp_hdr,
+                                       ne_duplicate_header, &areq->auth_hdr);
+        
 	
-    ne_add_response_header_handler(req, sess->spec->resp_info_hdr,
-				   ne_duplicate_header, 
-				   &areq->auth_info_hdr);
-    
-    sess->attempt = 0;
-
-    ne_set_request_private(req, sess->spec->id, areq);
+        ne_add_response_header_handler(req, sess->spec->resp_info_hdr,
+                                       ne_duplicate_header, 
+                                       &areq->auth_info_hdr);
+        
+        sess->attempt = 0;
+        
+        ne_set_request_private(req, sess->spec->id, areq);
+    }
 }
 
 
@@ -921,13 +1021,8 @@ static void ah_pre_send(ne_request *r, void *cookie, ne_buffer *request)
     auth_session *sess = cookie;
     struct auth_request *req = ne_get_request_private(r, sess->spec->id);
 
-    if (!sess->can_handle) {
+    if (!sess->can_handle || !req) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Not handling session.\n");
-    } else if (!is_in_domain(sess, req->uri)) {
-	/* We have moved out of the authentication domain */
-	NE_DEBUG(NE_DBG_HTTPAUTH, 
-	      "URI %s outside session domain, not handling.\n", req->uri);
-	req->will_handle = 0;
     } else {
 	char *value;
 
@@ -949,6 +1044,11 @@ static void ah_pre_send(ne_request *r, void *cookie, ne_buffer *request)
 	case auth_scheme_digest:
 	    value = request_digest(sess, req);
 	    break;
+#ifdef HAVE_GSSAPI	    
+	case auth_scheme_gssapi:
+	    value = request_gssapi(sess);
+	    break;
+#endif
 	default:
 	    value = NULL;
 	    break;
@@ -956,7 +1056,7 @@ static void ah_pre_send(ne_request *r, void *cookie, ne_buffer *request)
 
 	if (value != NULL) {
 	    ne_buffer_concat(request, sess->spec->req_hdr, ": ", value, NULL);
-	    free(value);
+	    ne_free(value);
 	}
 
     }
@@ -971,10 +1071,12 @@ static int ah_post_send(ne_request *req, void *cookie, const ne_status *status)
     struct auth_request *areq = ne_get_request_private(req, sess->spec->id);
     int ret = NE_OK;
 
+    if (!areq) return NE_OK;
+
     NE_DEBUG(NE_DBG_HTTPAUTH, 
-	  "ah_post_send (#%d), code is %d (want %d), %s is %s\n",
-	  sess->attempt, status->code, sess->spec->status_code, 
-	  sess->spec->resp_hdr, SAFELY(areq->auth_hdr));
+	     "ah_post_send (#%d), code is %d (want %d), %s is %s\n",
+	     sess->attempt, status->code, sess->spec->status_code, 
+	     sess->spec->resp_hdr, SAFELY(areq->auth_hdr));
     if (areq->auth_info_hdr != NULL && 
 	verify_response(areq, sess, areq->auth_info_hdr)) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Response authentication invalid.\n");
@@ -1001,7 +1103,7 @@ static void ah_destroy(ne_request *req, void *session)
 {
     auth_session *sess = session;
     struct auth_request *areq = ne_get_request_private(req, sess->spec->id);
-    free(areq);
+    if (areq) ne_free(areq);
 }
 
 static void free_auth(void *cookie)
@@ -1009,44 +1111,43 @@ static void free_auth(void *cookie)
     auth_session *sess = cookie;
 
     clean_session(sess);
-    free(sess);
+    ne_free(sess);
 }
 
-static void register_hooks(ne_session *sess, auth_session *as, const char *id)
-{
-    ne_hook_create_request(sess, ah_create, as);
-    ne_hook_pre_send(sess, ah_pre_send, as);
-    ne_hook_post_send(sess, ah_post_send, as);
-    ne_hook_destroy_request(sess, ah_destroy, as);
-    ne_hook_destroy_session(sess, free_auth, as);
-
-    ne_set_session_private(sess, id, as);
-}
-
-static void auth_register(ne_session *sess,
+static void auth_register(ne_session *sess, int isproxy,
 			  const struct auth_class *ahc, const char *id, 
-			  ne_request_auth callback, void *userdata) 
+			  ne_auth_creds creds, void *userdata) 
 {
     auth_session *ahs = ne_calloc(sizeof *ahs);
 
-    ahs->reqcreds = callback;
-    ahs->reqcreds_udata = userdata;
+    ahs->creds = creds;
+    ahs->userdata = userdata;
     ahs->sess = sess;
     ahs->spec = ahc;
 
-    register_hooks(sess, ahs, id);
+    if (strcmp(ne_get_scheme(sess), "https") == 0)
+        ahs->context = isproxy ? AUTH_CONNECT : AUTH_NOTCONNECT;
+    else
+        ahs->context = AUTH_ANY;
+    
+    /* Register hooks */
+    ne_hook_create_request(sess, ah_create, ahs);
+    ne_hook_pre_send(sess, ah_pre_send, ahs);
+    ne_hook_post_send(sess, ah_post_send, ahs);
+    ne_hook_destroy_request(sess, ah_destroy, ahs);
+    ne_hook_destroy_session(sess, free_auth, ahs);
+
+    ne_set_session_private(sess, id, ahs);
 }
 
-void ne_set_server_auth(ne_session *sess, ne_request_auth callback, 
-			void *userdata)
+void ne_set_server_auth(ne_session *sess, ne_auth_creds creds, void *userdata)
 {
-    auth_register(sess, &ah_server_class, HOOK_SERVER_ID, callback, userdata);
+    auth_register(sess, 0, &ah_server_class, HOOK_SERVER_ID, creds, userdata);
 }
 
-void ne_set_proxy_auth(ne_session *sess, ne_request_auth callback, 
-		       void *userdata)
+void ne_set_proxy_auth(ne_session *sess, ne_auth_creds creds, void *userdata)
 {
-    auth_register(sess, &ah_proxy_class, HOOK_PROXY_ID, callback, userdata);
+    auth_register(sess, 1, &ah_proxy_class, HOOK_PROXY_ID, creds, userdata);
 }
 
 void ne_forget_auth(ne_session *sess)
