@@ -1,6 +1,6 @@
 /* 
    socket handling routines
-   Copyright (C) 1998-2002, Joe Orton <joe@manyfish.co.uk>, 
+   Copyright (C) 1998-2003, Joe Orton <joe@manyfish.co.uk>, 
    Copyright (C) 1999-2000 Tommi Komulainen <Tommi.Komulainen@iki.fi>
 
    This library is free software; you can redistribute it and/or
@@ -26,6 +26,15 @@
 
 #include "config.h"
 
+#ifdef __hpux
+/* pick up hstrerror */
+#define _XOPEN_SOURCE_EXTENDED 1
+/* don't use the broken getaddrinfo shipped in HP-UX 11.11 */
+#ifdef USE_GETADDRINFO
+#undef USE_GETADDRINFO
+#endif
+#endif
+
 #include <sys/types.h>
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
@@ -41,16 +50,19 @@
 #ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
 #endif
+#ifdef HAVE_NETINET_TCP_H
+#include <netinet/tcp.h>
+#endif
 #ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
 #endif
+#ifdef HAVE_NETDB_H
+#include <netdb.h>
+#endif
 
 #ifdef WIN32
 #include <winsock2.h>
 #include <stddef.h>
-#else
-#include <netinet/in.h>
-#include <netdb.h>
 #endif
 
 #if defined(NEON_SSL) && defined(HAVE_LIMITS_H)
@@ -79,6 +91,15 @@
 #include <socks.h>
 #endif
 
+#ifdef NEON_SSL
+#include <openssl/ssl.h>
+#include <openssl/err.h>
+#include <openssl/pkcs12.h> /* for PKCS12_PBE_add */
+#include <openssl/rand.h>
+
+#include "ne_privssl.h"
+#endif
+
 #include "ne_i18n.h"
 #include "ne_utils.h"
 #include "ne_string.h"
@@ -91,6 +112,14 @@
  * structure, or losing type-safety by using void *. */
 #ifdef USE_GETADDRINFO
 typedef struct addrinfo ne_inet_addr;
+
+/* To avoid doing AAAA queries unless absolutely necessary, either use
+ * AI_ADDRCONFIG where available, or a run-time check for working IPv6
+ * support; the latter is only known to work on Linux. */
+#if !defined(USE_GAI_ADDRCONFIG) && defined(__linux__)
+#define USE_CHECK_IPV6
+#endif
+
 #else
 typedef struct in_addr ne_inet_addr;
 #endif
@@ -100,28 +129,31 @@ typedef struct in_addr ne_inet_addr;
 
 #if defined(__BEOS__) && !defined(BONE_VERSION)
 /* pre-BONE */
-#define NEON_WRITE(a,b,c) send(a,b,c,0)
-#define NEON_READ(a,b,c) recv(a,b,c,0)
-#define NEON_CLOSE(s) closesocket(s)
+#define ne_write(a,b,c) send(a,b,c,0)
+#define ne_read(a,b,c) recv(a,b,c,0)
+#define ne_close(s) closesocket(s)
+#define ne_errno errno
 #elif defined(WIN32)
-#define NEON_WRITE(a,b,c) send(a,b,c,0)
-#define NEON_READ(a,b,c) recv(a,b,c,0)
-#define NEON_CLOSE(s) closesocket(s)
+#define ne_write(a,b,c) send(a,b,c,0)
+#define ne_read(a,b,c) recv(a,b,c,0)
+#define ne_close(s) closesocket(s)
+#define ne_errno WSAGetLastError()
 #else /* really Unix! */
-#define NEON_WRITE(a,b,c) write(a,b,c)
-#define NEON_READ(a,b,c) read(a,b,c)
-#define NEON_CLOSE(s) close(s)
+#define ne_write(a,b,c) write(a,b,c)
+#define ne_read(a,b,c) read(a,b,c)
+#define ne_close(s) close(s)
+#define ne_errno errno
 #endif
 
-#if !defined(ECONNRESET) && defined(WIN32)
-#define ECONNRESET WSAECONNRESET
-#endif
-
-#ifdef NEON_SSL
-#include <openssl/ssl.h>
-#include <openssl/err.h>
-#include <openssl/pkcs12.h> /* for PKCS12_PBE_add */
-#include <openssl/rand.h>
+#ifdef WIN32
+#define NE_ISRESET(e) ((e) == WSAECONNABORTED || (e) == WSAETIMEDOUT || \
+                       (e) == WSAECONNRESET || (e) == WSAENETRESET)
+#define NE_ISCLOSED(e) ((e) == WSAESHUTDOWN || (e) == WSAENOTCONN)
+#define NE_ISINTR(e) (0)
+#else /* Unix */
+#define NE_ISRESET(e) ((e) == ECONNRESET)
+#define NE_ISCLOSED(e) ((e) == EPIPE)
+#define NE_ISINTR(e) ((e) == EINTR)
 #endif
 
 /* Socket read timeout */
@@ -144,13 +176,11 @@ struct iofns {
 struct ne_socket_s {
     int fd;
     char error[200];
-    ne_progress progress_cb;
     void *progress_ud;
     int rdtimeout; /* read timeout. */
     const struct iofns *ops;
 #ifdef NEON_SSL
-    SSL *ssl;
-    SSL_CTX *ssl_ctx;
+    ne_ssl_socket ssl;
 #endif
     /* The read buffer: ->buffer stores byte which have been read; as
      * these are consumed and passed back to the caller, bufpos
@@ -177,14 +207,21 @@ struct ne_sock_addr_s {
 /* set_error: set socket error string to 'str'. */
 #define set_error(s, str) ne_strnzcpy((s)->error, (str), sizeof (s)->error)
 
-/* set_strerror: set socket error to system error string for 'e' */
-#define set_strerror(s, e) ne_strerror((e), (s)->error, sizeof (s)->error)
-
-void ne_register_progress(ne_socket *sock, ne_progress cb, void *userdata)
+/* set_strerror: set socket error to system error string for 'errnum' */
+#ifdef WIN32
+/* Print system error message to given buffer. */
+static void print_error(int errnum, char *buffer, size_t buflen)
 {
-    sock->progress_cb = cb;
-    sock->progress_ud = userdata;
+    if (FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM
+                       | FORMAT_MESSAGE_IGNORE_INSERTS,
+                       NULL, (DWORD) errnum, 0, 
+                       buffer, buflen, NULL) == 0)
+        ne_snprintf(buffer, buflen, "Socket error %d", errnum);
 }
+#define set_strerror(s, e) print_error((e), (s)->error, sizeof (s)->error)
+#else /* not WIN32 */
+#define set_strerror(s, e) ne_strerror((e), (s)->error, sizeof (s)->error)
+#endif
 
 #ifdef NEON_SSL
 static int prng_seeded = 0;
@@ -192,34 +229,57 @@ static int prng_seeded = 0;
 /* Initialize SSL library; returns non-zero on failure. */
 static int init_ssl(void)
 {
-    const char *file;
-
     SSL_load_error_strings();
     SSL_library_init();
     PKCS12_PBE_add();  /* ### not sure why this is needed. */
 
+    /* Check whether the PRNG has already been seeded. */
     if (RAND_status() == 1)
-	return 0; /* already seeded internally by OpenSSL */
-
-    NE_DEBUG(NE_DBG_SOCKET, "Trying to seed PRNG...");
-
-    file = getenv("EGDSOCKET");
-    if (file && RAND_egd(file) != -1)
 	return 0;
 
-    file = getenv("HOME");
-    if (file) {
-	char path[1024];    
-	ne_snprintf(path, sizeof path, "%s/.entropy", file);
-	if (RAND_egd(path) != -1)
-	    return 0;
+#ifdef EGD_PATH
+    NE_DEBUG(NE_DBG_SOCKET, "Seeding PRNG from " EGD_PATH "...\n");
+    if (RAND_egd(EGD_PATH) != -1)
+	return 0;
+#elif defined(ENABLE_EGD)
+    {
+	static const char *paths[] = { "/var/run/egd-pool", "/dev/egd-pool",
+				       "/etc/egd-pool", "/etc/entropy" };
+	size_t n;
+	for (n = 0; n < sizeof(paths) / sizeof(char *); n++) {
+	    NE_DEBUG(NE_DBG_SOCKET, "Seeding PRNG from %s...\n", paths[n]);
+	    if (RAND_egd(paths[n]) != -1)
+		return 0;
+	}
     }
+#endif /* EGD_PATH */
 
     NE_DEBUG(NE_DBG_SOCKET, "No entropy source found; could not seed PRNG.\n");
     return -1;
 }
 #endif /* NEON_SSL */
 
+#ifdef USE_CHECK_IPV6
+static int ipv6_disabled = 0;
+
+/* On Linux kernels, IPv6 is typically built as a loadable module, and
+ * socket(AF_INET6, ...) will fail if this module is not loaded, so
+ * the slow AAAA lookups can be avoided for this common case. */
+static void init_ipv6(void)
+{
+    int fd = socket(AF_INET6, SOCK_STREAM, 0);
+    
+    if (fd < 0)
+        ipv6_disabled = 1;
+    else
+        close(fd);
+}
+#else
+#define ipv6_disabled (0)
+#endif
+
+static int init_result = 0;
+
 int ne_sock_init(void)
 {
 #ifdef WIN32
@@ -227,11 +287,10 @@ int ne_sock_init(void)
     WSADATA wsaData;
     int err;
 #endif
-    static int result = 0; 
 
-    if (result > 0) 
+    if (init_result > 0) 
 	return 0;
-    else if (result < 0)
+    else if (init_result < 0)
 	return -1;
 
 #ifdef WIN32    
@@ -239,7 +298,7 @@ int ne_sock_init(void)
     
     err = WSAStartup(wVersionRequested, &wsaData);
     if (err != 0) {
-	result = -1;
+	init_result = -1;
 	return -1;
     }
 
@@ -253,16 +312,19 @@ int ne_sock_init(void)
     (void) signal(SIGPIPE, SIG_IGN);
 #endif
 
-#ifdef NEON_SSL
-    if (init_ssl()) {
-	NE_DEBUG(NE_DBG_SOCKET, "SSL initialization failed; lacking PRNG?");
-	result = -1;
-	return -1;
-    }
-    prng_seeded = 1;
+#ifdef USE_CHECK_IPV6
+    init_ipv6();
 #endif
 
-    result = 1;
+#ifdef NEON_SSL
+    if (init_ssl()) {
+	NE_DEBUG(NE_DBG_SOCKET, "SSL initialization failed; lacking PRNG?\n");
+    } else {
+        prng_seeded = 1;
+    }
+#endif
+
+    init_result = 1;
     return 0;
 }
 
@@ -270,7 +332,8 @@ void ne_sock_exit(void)
 {
 #ifdef WIN32
     WSACleanup();
-#endif	
+#endif
+    init_result = 0;
 }
 
 int ne_sock_block(ne_socket *sock, int n)
@@ -280,6 +343,9 @@ int ne_sock_block(ne_socket *sock, int n)
     return sock->ops->readable(sock, n);
 }
 
+/* Cast address object AD to type 'sockaddr_TY' */ 
+#define SACAST(ty, ad) ((struct sockaddr_##ty *)(ad))
+
 #define SOCK_ERR(x) do { ssize_t _sock_err = (x); \
 if (_sock_err < 0) return _sock_err; } while(0)
 
@@ -358,11 +424,10 @@ static int readable_raw(ne_socket *sock, int secs)
 	    tvp->tv_sec = secs;
 	    tvp->tv_usec = 0;
 	}
-	errno = 0;
 	ret = select(fdno + 1, &rdfds, NULL, NULL, tvp);
-    } while (ret < 0 && errno == EINTR);
+    } while (ret < 0 && NE_ISINTR(ne_errno));
     if (ret < 0) {
-	set_strerror(sock, errno);
+	set_strerror(sock, ne_errno);
 	return NE_SOCK_ERROR;
     }
     return (ret == 0) ? NE_SOCK_TIMEOUT : 0;
@@ -376,36 +441,38 @@ static ssize_t read_raw(ne_socket *sock, char *buffer, size_t len)
     if (ret) return ret;
 
     do {
-	ret = NEON_READ(sock->fd, buffer, len);
-    } while (ret == -1 && errno == EINTR);
+	ret = ne_read(sock->fd, buffer, len);
+    } while (ret == -1 && NE_ISINTR(ne_errno));
 
     if (ret == 0) {
 	set_error(sock, _("Connection closed"));
 	ret = NE_SOCK_CLOSED;
     } else if (ret < 0) {
-	int errnum = errno;
-	ret = (errnum == ECONNRESET) ? NE_SOCK_RESET : NE_SOCK_ERROR;
+	int errnum = ne_errno;
+	ret = NE_ISRESET(errnum) ? NE_SOCK_RESET : NE_SOCK_ERROR;
 	set_strerror(sock, errnum);
     }
 
     return ret;
 }
 
-#define MAP_ERR(e) ((e) == EPIPE ? NE_SOCK_CLOSED : \
-((e) == ECONNRESET ? NE_SOCK_RESET : NE_SOCK_ERROR))
+#define MAP_ERR(e) (NE_ISCLOSED(e) ? NE_SOCK_CLOSED : \
+                    (NE_ISRESET(e) ? NE_SOCK_RESET : NE_SOCK_ERROR))
 
 static ssize_t write_raw(ne_socket *sock, const char *data, size_t length) 
 {
     ssize_t wrote;
     
     do {
-	wrote = NEON_WRITE(sock->fd, data, length);
-	data += wrote;
-	length -= wrote;
-    } while ((wrote > 0 || errno == EINTR) && length > 0);
+	wrote = ne_write(sock->fd, data, length);
+        if (wrote > 0) {
+            data += wrote;
+            length -= wrote;
+        }
+    } while ((wrote > 0 || NE_ISINTR(ne_errno)) && length > 0);
 
     if (wrote < 0) {
-	int errnum = errno;
+	int errnum = ne_errno;
 	set_strerror(sock, errnum);
 	return MAP_ERR(errnum);
     }
@@ -424,7 +491,7 @@ static int readable_ossl(ne_socket *sock, int secs)
      * SSL_pending returns non-zero.  Possibly need to do
      * SSL_read(ssl, buf, SSL_pending(ssl)) */
 
-    if (sock->ssl && SSL_pending(sock->ssl))
+    if (SSL_pending(sock->ssl.ssl))
 	return 0;
 
     return readable_raw(sock, secs);
@@ -433,7 +500,7 @@ static int readable_ossl(ne_socket *sock, int secs)
 /* SSL error handling, according to SSL_get_error(3). */
 static int error_ossl(ne_socket *sock, int sret)
 {
-    int err = SSL_get_error(sock->ssl, sret), ret = NE_SOCK_ERROR;
+    int err = SSL_get_error(sock->ssl.ssl, sret), ret = NE_SOCK_ERROR;
     
     switch (err) {
     case SSL_ERROR_ZERO_RETURN:
@@ -449,7 +516,7 @@ static int error_ossl(ne_socket *sock, int sret)
 		ret = NE_SOCK_TRUNC;
 	    } else {
 		/* Other socket error. */
-		err = errno;
+		err = ne_errno;
 		set_strerror(sock, err);
 		ret = MAP_ERR(err);
 	    }
@@ -477,7 +544,7 @@ static ssize_t read_ossl(ne_socket *sock, char *buffer, size_t len)
     ret = readable_ossl(sock, sock->rdtimeout);
     if (ret) return ret;
     
-    ret = SSL_read(sock->ssl, buffer, CAST2INT(len));
+    ret = SSL_read(sock->ssl.ssl, buffer, CAST2INT(len));
     if (ret <= 0)
 	ret = error_ossl(sock, ret);
 
@@ -487,7 +554,7 @@ static ssize_t read_ossl(ne_socket *sock, char *buffer, size_t len)
 static ssize_t write_ossl(ne_socket *sock, const char *data, size_t len)
 {
     int ret, ilen = CAST2INT(len);
-    ret = SSL_write(sock->ssl, data, ilen);
+    ret = SSL_write(sock->ssl.ssl, data, ilen);
     /* ssl.h says SSL_MODE_ENABLE_PARTIAL_WRITE must be enabled to
      * have SSL_write return < length...  so, SSL_write should never
      * return < length. */
@@ -570,6 +637,11 @@ ssize_t ne_sock_fullread(ne_socket *sock, char *buffer, size_t buflen)
 #define INADDR_NONE ((unsigned long) -1)
 #endif
 
+#if !defined(USE_GETADDRINFO) && !defined(HAVE_DECL_H_ERRNO) && !defined(WIN32)
+/* Ancient versions of netdb.h don't export h_errno. */
+extern int h_errno;
+#endif
+
 /* This implemementation does not attempt to support IPv6 using
  * gethostbyname2 et al.  */
 ne_sock_addr *ne_addr_resolve(const char *hostname, int flags)
@@ -578,18 +650,25 @@ ne_sock_addr *ne_addr_resolve(const char *hostname, int flags)
 #ifdef USE_GETADDRINFO
     struct addrinfo hints = {0};
     char *pnt;
-    hints.ai_family = AF_UNSPEC;
     hints.ai_socktype = SOCK_STREAM;
     if (hostname[0] == '[' && ((pnt = strchr(hostname, ']')) != NULL)) {
 	char *hn = ne_strdup(hostname + 1);
 	hn[pnt - hostname - 1] = '\0';
-#ifdef AI_NUMERICHOST /* this was missing from the RFC2133 API */
+#ifdef AI_NUMERICHOST /* added in the RFC2553 API */
 	hints.ai_flags = AI_NUMERICHOST;
 #endif
+        hints.ai_family = AF_INET6;
 	addr->errnum = getaddrinfo(hn, NULL, &hints, &addr->result);
-	free(hn);
+	ne_free(hn);
     } else {
+#ifdef USE_GAI_ADDRCONFIG /* added in the RFC3493 API */
+        hints.ai_flags = AI_ADDRCONFIG;
+        hints.ai_family = AF_UNSPEC;
+        addr->errnum = getaddrinfo(hostname, NULL, &hints, &addr->result);
+#else
+        hints.ai_family = ipv6_disabled ? AF_INET : AF_UNSPEC;
 	addr->errnum = getaddrinfo(hostname, NULL, &hints, &addr->result);
+#endif
     }
 #else /* Use gethostbyname() */
     unsigned long laddr;
@@ -599,7 +678,11 @@ ne_sock_addr *ne_addr_resolve(const char *hostname, int flags)
     if (laddr == INADDR_NONE) {
 	hp = gethostbyname(hostname);
 	if (hp == NULL) {
-	    addr->errnum = h_errno;
+#ifdef WIN32
+	    addr->errnum = WSAGetLastError();
+#else
+            addr->errnum = h_errno;
+#endif
 	} else if (hp->h_length != sizeof(struct in_addr)) {
 	    /* fail gracefully if somebody set RES_USE_INET6 */
 	    addr->errnum = NO_RECOVERY;
@@ -629,7 +712,7 @@ int ne_addr_result(const ne_sock_addr *addr)
     return addr->errnum;
 }
 
-ne_inet_addr *ne_addr_first(ne_sock_addr *addr)
+const ne_inet_addr *ne_addr_first(ne_sock_addr *addr)
 {
 #ifdef USE_GETADDRINFO
     addr->cursor = addr->result->ai_next;
@@ -640,7 +723,7 @@ ne_inet_addr *ne_addr_first(ne_sock_addr *addr)
 #endif
 }
 
-ne_inet_addr *ne_addr_next(ne_sock_addr *addr)
+const ne_inet_addr *ne_addr_next(ne_sock_addr *addr)
 {
 #ifdef USE_GETADDRINFO
     struct addrinfo *ret = addr->cursor;
@@ -657,9 +740,12 @@ ne_inet_addr *ne_addr_next(ne_sock_addr *addr)
 
 char *ne_addr_error(const ne_sock_addr *addr, char *buf, size_t bufsiz)
 {
+#ifdef WIN32
+    print_error(addr->errnum, buf, bufsiz);
+#else
     const char *err;
 #ifdef USE_GETADDRINFO
-    /* over-ride horrible generic "Name or service not known" error. */
+    /* override horrible generic "Name or service not known" error. */
     if (addr->errnum == EAI_NONAME)
 	err = _("Host not found");
     else
@@ -670,21 +756,22 @@ char *ne_addr_error(const ne_sock_addr *addr, char *buf, size_t bufsiz)
     err = _("Host not found");
 #endif
     ne_strnzcpy(buf, err, bufsiz);
+#endif /* WIN32 */
     return buf;
 }
 
-char *ne_addr_print(const ne_inet_addr *ia, char *buf, size_t bufsiz)
+char *ne_iaddr_print(const ne_inet_addr *ia, char *buf, size_t bufsiz)
 {
 #ifdef USE_GETADDRINFO /* implies inet_ntop */
     const char *ret;
 #ifdef AF_INET6
     if (ia->ai_family == AF_INET6) {
-	struct sockaddr_in6 *in6 = (struct sockaddr_in6 *)ia->ai_addr;
+	struct sockaddr_in6 *in6 = SACAST(in6, ia->ai_addr);
 	ret = inet_ntop(AF_INET6, &in6->sin6_addr, buf, bufsiz);
     } else
 #endif
     if (ia->ai_family == AF_INET) {
-	struct sockaddr_in *in = (struct sockaddr_in *)ia->ai_addr;
+	struct sockaddr_in *in = SACAST(in, ia->ai_addr);
 	ret = inet_ntop(AF_INET, &in->sin_addr, buf, bufsiz);
     } else
 	ret = NULL;
@@ -703,26 +790,35 @@ void ne_addr_destroy(ne_sock_addr *addr)
 	freeaddrinfo(addr->result);
 #else
     if (addr->addrs)
-	free(addr->addrs);
+	ne_free(addr->addrs);
 #endif
-    free(addr);
+    ne_free(addr);
 }
 
 /* Connect socket 'fd' to address 'addr' on given 'port': */
-static int raw_connect(int fd, ne_inet_addr *addr, unsigned int port)
+static int raw_connect(int fd, const ne_inet_addr *addr, unsigned int port)
 {
 #ifdef USE_GETADDRINFO
 #ifdef AF_INET6
+    /* fill in the _family field for AIX 4.3, which forgets to do so. */
     if (addr->ai_family == AF_INET6) {
-	struct sockaddr_in6 *in6 = (struct sockaddr_in6 *)addr->ai_addr;
-	in6->sin6_port = port;
+	struct sockaddr_in6 in6;
+	memcpy(&in6, addr->ai_addr, sizeof in6);
+	in6.sin6_port = port;
+        in6.sin6_family = AF_INET6;
+	return connect(fd, (struct sockaddr *)&in6, sizeof in6);
     } else
 #endif
     if (addr->ai_family == AF_INET) {
-	struct sockaddr_in *in = (struct sockaddr_in *)addr->ai_addr;
-	in->sin_port = port;
-    } /* else something weird, and the connect should fail... */
-    return connect(fd, addr->ai_addr, addr->ai_addrlen);
+	struct sockaddr_in in;
+	memcpy(&in, addr->ai_addr, sizeof in);
+	in.sin_port = port;
+        in.sin_family = AF_INET;
+	return connect(fd, (struct sockaddr *)&in, sizeof in);
+    } else {
+	errno = EINVAL;
+	return -1;
+    }
 #else
     struct sockaddr_in sa = {0};
     sa.sin_family = AF_INET;
@@ -732,17 +828,18 @@ static int raw_connect(int fd, ne_inet_addr *addr, unsigned int port)
 #endif
 }
 
-static ne_socket *create_sock(int fd)
+ne_socket *ne_sock_create(void)
 {
     ne_socket *sock = ne_calloc(sizeof *sock);
-    sock->fd = fd;
     sock->rdtimeout = SOCKET_READ_TIMEOUT;
     sock->bufpos = sock->buffer;
     sock->ops = &iofns_raw;
+    sock->fd = -1;
     return sock;
 }
 
-ne_socket *ne_sock_connect(ne_inet_addr *addr, unsigned int portnum)
+int ne_sock_connect(ne_socket *sock,
+                    const ne_inet_addr *addr, unsigned int port)
 {
     int fd;
 
@@ -753,23 +850,104 @@ ne_socket *ne_sock_connect(ne_inet_addr *addr, unsigned int portnum)
 #else
     fd = socket(AF_INET, SOCK_STREAM, 0);
 #endif
-    if (fd < 0)
-	return NULL;
+    if (fd < 0) {
+        set_strerror(sock, ne_errno);
+	return -1;
+    }
     
-    if (raw_connect(fd, addr, ntohs(portnum))) {
-	NEON_CLOSE(fd);
-	return NULL;
+#if defined(TCP_NODELAY) && defined(HAVE_SETSOCKOPT) && defined(IPPROTO_TCP)
+    { /* Disable the Nagle algorithm; better to add write buffering
+       * instead of doing this. */
+        int flag = 1;
+        setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof flag);
+    }
+#endif
+
+    if (raw_connect(fd, addr, ntohs(port))) {
+        set_strerror(sock, ne_errno);
+	ne_close(fd);
+	return -1;
     }
 
-    return create_sock(fd);
+    sock->fd = fd;
+    return 0;
 }
 
-ne_socket *ne_sock_accept(int listener) 
+ne_inet_addr *ne_iaddr_make(ne_iaddr_type type, const unsigned char *raw)
+{
+    ne_inet_addr *ia;
+#if !defined(AF_INET6) || !defined(USE_GETADDRINFO)
+    /* fail if IPv6 address is given if IPv6 is not supported. */
+    if (type == ne_iaddr_ipv6)
+	return NULL;
+#endif
+    ia = ne_calloc(sizeof *ia);
+#ifdef USE_GETADDRINFO
+    /* ai_protocol and ai_socktype aren't used by raw_connect so
+     * ignore them here. (for now) */
+    if (type == ne_iaddr_ipv4) {
+	struct sockaddr_in *in4 = ne_calloc(sizeof *in4);
+	ia->ai_family = AF_INET;
+	ia->ai_addr = (struct sockaddr *)in4;
+	ia->ai_addrlen = sizeof *in4;
+	in4->sin_family = AF_INET;
+	memcpy(&in4->sin_addr.s_addr, raw, sizeof in4->sin_addr.s_addr);
+    }
+#ifdef AF_INET6
+    else {
+	struct sockaddr_in6 *in6 = ne_calloc(sizeof *in6);
+	ia->ai_family = AF_INET6;
+	ia->ai_addr = (struct sockaddr *)in6;
+	ia->ai_addrlen = sizeof *in6;
+	in6->sin6_family = AF_INET6;
+	memcpy(&in6->sin6_addr, raw, sizeof in6->sin6_addr.s6_addr);
+    }
+#endif
+#else /* !USE_GETADDRINFO */
+    memcpy(&ia->s_addr, raw, sizeof ia->s_addr);
+#endif    
+    return ia;
+}
+
+int ne_iaddr_cmp(const ne_inet_addr *i1, const ne_inet_addr *i2)
+{
+#ifdef USE_GETADDRINFO
+    if (i1->ai_family != i2->ai_family)
+	return i2->ai_family - i1->ai_family;
+    if (i1->ai_family == AF_INET) {
+	struct sockaddr_in *in1 = SACAST(in, i1->ai_addr), 
+	    *in2 = SACAST(in, i2->ai_addr);
+	return memcmp(&in1->sin_addr.s_addr, &in2->sin_addr.s_addr, 
+		      sizeof in1->sin_addr.s_addr);
+    } else if (i1->ai_family == AF_INET6) {
+	struct sockaddr_in6 *in1 = SACAST(in6, i1->ai_addr), 
+	    *in2 = SACAST(in6, i2->ai_addr);
+	return memcmp(in1->sin6_addr.s6_addr, in2->sin6_addr.s6_addr,
+		      sizeof in1->sin6_addr.s6_addr);
+    } else
+	return -1;
+#else
+    return memcmp(&i1->s_addr, &i2->s_addr, sizeof i1->s_addr);
+#endif
+}
+
+void ne_iaddr_free(ne_inet_addr *addr)
+{
+#ifdef USE_GETADDRINFO
+    ne_free(addr->ai_addr);
+#endif
+    ne_free(addr);
+}
+
+int ne_sock_accept(ne_socket *sock, int listener) 
 {
     int fd = accept(listener, NULL, NULL);
-    if (fd > 0)
-	return create_sock(fd);
-    return NULL;
+
+    if (fd < 0)
+        return -1;
+
+    sock->fd = fd;
+    return 0;
 }
 
 int ne_sock_fd(const ne_socket *sock)
@@ -783,28 +961,16 @@ void ne_sock_read_timeout(ne_socket *sock, int timeout)
 }
 
 #ifdef NEON_SSL
-int ne_sock_use_ssl(ne_socket *sock)
-{
-    /* create a context. */
-    sock->ssl_ctx = SSL_CTX_new(SSLv23_client_method());
-    return ne_sock_use_ssl_os(sock, sock->ssl_ctx, NULL, NULL, NULL);
-}
 
-void ne_sock_switch_ssl(ne_socket *sock, SSL *ssl)
+void ne_sock_switch_ssl(ne_socket *sock, void *ssl)
 {
-    sock->ssl = ssl;
+    sock->ssl.ssl = ssl;
     sock->ops = &iofns_ossl;
 }
 
-#else
-int ne_sock_use_ssl(ne_socket *sock) { return -1; }
-#endif
-
-#ifdef NEON_SSL
-int ne_sock_use_ssl_os(ne_socket *sock, SSL_CTX *ctx, 
-		       SSL_SESSION *sess, SSL **ssl,
-		       void *appdata)
+int ne_sock_connect_ssl(ne_socket *sock, ne_ssl_context *ctx)
 {
+    SSL *ssl;
     int ret;
 
     if (!prng_seeded && RAND_status() != 1) {
@@ -812,39 +978,36 @@ int ne_sock_use_ssl_os(ne_socket *sock, SSL_CTX *ctx,
 	return NE_SOCK_ERROR;
     }
 
-    sock->ssl = SSL_new(ctx);
-    if (!sock->ssl) {
+    sock->ssl.ssl = ssl = SSL_new(ctx->ctx);
+    if (!ssl) {
 	set_error(sock, _("Could not create SSL structure"));
 	return NE_SOCK_ERROR;
     }
     
-    if (appdata) {
-	SSL_set_app_data(sock->ssl, appdata);
-    }
-
+    SSL_set_app_data(ssl, ctx);
+    SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);
+    SSL_set_fd(ssl, sock->fd);
     sock->ops = &iofns_ossl;
-
-    SSL_set_mode(sock->ssl, SSL_MODE_AUTO_RETRY);
-
-    SSL_set_fd(sock->ssl, sock->fd);
     
-    if (sess)
-	SSL_set_session(sock->ssl, sess);
+    if (ctx->sess)
+	SSL_set_session(ssl, ctx->sess);
 
-    ret = SSL_connect(sock->ssl);
+    ret = SSL_connect(ssl);
     if (ret != 1) {
 	error_ossl(sock, ret);
-	SSL_free(sock->ssl);
-	sock->ssl = NULL;
+	SSL_free(ssl);
+	sock->ssl.ssl = NULL;
 	return NE_SOCK_ERROR;
     }
 
-    if (ssl)
-	*ssl = sock->ssl;
-
     return 0;
 }
 
+ne_ssl_socket *ne_sock_sslsock(ne_socket *sock)
+{
+    return &sock->ssl;
+}
+
 #endif
 
 const char *ne_sock_error(const ne_socket *sock)
@@ -857,15 +1020,16 @@ int ne_sock_close(ne_socket *sock)
 {
     int ret;
 #ifdef NEON_SSL
-    if (sock->ssl) {
-	SSL_shutdown(sock->ssl);
-	SSL_free(sock->ssl);
-	if (sock->ssl_ctx)
-	    SSL_CTX_free(sock->ssl_ctx);
+    if (sock->ssl.ssl) {
+	SSL_shutdown(sock->ssl.ssl);
+	SSL_free(sock->ssl.ssl);
     }
 #endif
-    ret = NEON_CLOSE(sock->fd);
-    free(sock);
+    if (sock->fd < 0)
+        ret = 0;
+    else
+        ret = ne_close(sock->fd);
+    ne_free(sock);
     return ret;
 }
 
