@@ -7,6 +7,9 @@ OPENSSL=${OPENSSL-openssl}
 CONF=${srcdir}/openssl.conf
 REQ="${OPENSSL} req -config ${CONF}"
 CA="${OPENSSL} ca -config ${CONF} -batch"
+# MKCERT makes a self-signed cert
+MKCERT="${REQ} -x509 -new -days 900"
+
 REQDN=reqDN
 export REQDN
 
@@ -19,7 +22,7 @@ echo 01 > ca/serial
 ${OPENSSL} genrsa -rand ${srcdir}/../configure > ca/key.pem
 ${OPENSSL} genrsa -rand ${srcdir}/../configure > client.key
 
-${REQ} -x509 -new -key ca/key.pem -out ca/cert.pem <<EOF
+${MKCERT} -key ca/key.pem -out ca/cert.pem <<EOF
 US
 California
 Oakland
@@ -31,13 +34,16 @@ EOF
 
 # Function to generate appropriate output for `openssl req'.
 csr_fields() {
-OU=${1-"Neon QA Dept"}
 CN=${2-"localhost"}
+OU=${1-"Neon QA Dept"}
+Org=${3-"Neon Hackers Ltd"}
+Locality=${4-"Cambridge"}
+State=${5-"Cambridgeshire"}
 cat <<EOF
 GB
-Cambridgeshire
-Cambridge
-Neon Hackers Ltd
+${State}
+${Locality}
+${Org}
 ${OU}
 ${CN}
 neon@webdav.org
@@ -52,7 +58,7 @@ csr_fields "Upper Case Dept" lOcALhost | \
 ${REQ} -new -key ${srcdir}/server.key -out caseless.csr
 
 csr_fields "Use AltName Dept" nowhere.example.com | \
-${REQ} -new -key ${srcdir}/server.key -out altname.csr
+${REQ} -new -key ${srcdir}/server.key -out altname1.csr
 
 csr_fields "Two AltName Dept" nowhere.example.com | \
 ${REQ} -new -key ${srcdir}/server.key -out altname2.csr
@@ -63,11 +69,30 @@ ${REQ} -new -key ${srcdir}/server.key -out altname3.csr
 csr_fields "Fourth AltName Dept" localhost | \
 ${REQ} -new -key ${srcdir}/server.key -out altname4.csr
 
+csr_fields "Fifth Altname Dept" localhost | \
+${REQ} -new -key ${srcdir}/server.key -out altname5.csr
+
 csr_fields "Self-Signed" | \
-${REQ} -new -x509 -key ${srcdir}/server.key -out ssigned.pem
+${MKCERT} -key ${srcdir}/server.key -out ssigned.pem
 
 csr_fields "Bad Hostname Department" nohost.example.com | \
-${REQ} -new -x509 -key ${srcdir}/server.key -out wrongcn.pem
+${MKCERT} -key ${srcdir}/server.key -out wrongcn.pem
+
+### produce a set of CA certs
+
+csr_fields "First Random CA" "first.example.com" "CAs Ltd." Lincoln Lincolnshire | \
+${MKCERT} -key ${srcdir}/server.key -out ca1.pem
+
+csr_fields "Second Random CA" "second.example.com" "CAs Ltd." Falmouth Cornwall | \
+${MKCERT} -key ${srcdir}/server.key -out ca2.pem
+
+csr_fields "Third Random CA" "third.example.com" "CAs Ltd." Ipswich Suffolk | \
+${MKCERT} -key ${srcdir}/server.key -out ca3.pem
+
+csr_fields "Fourth Random CA" "fourth.example.com" "CAs Ltd." Norwich Norfolk | \
+${MKCERT} -key ${srcdir}/server.key -out ca4.pem
+
+cat ca[1234].pem > calist.pem
 
 # Only works with a Linuxy hostname command: continue without it,
 # as appropriate tests are skipped if these fail.
@@ -83,20 +108,44 @@ fi
 csr_fields "Neon Client Cert" ignored.example.com | \
 ${REQ} -new -key client.key -out client.csr
 
+### requests using special DN.
+
 REQDN=reqDN.doubleCN
 csr_fields "Double CN Dept" "nohost.example.com
 localhost" | ${REQ} -new -key ${srcdir}/server.key -out twocn.csr
 
-for f in server client twocn caseless; do
+REQDN=reqDN.CNfirst
+echo localhost | ${REQ} -new -key ${srcdir}/server.key -out cnfirst.csr
+
+REQDN=reqDN.missingCN
+echo GB | ${REQ} -new -key ${srcdir}/server.key -out missingcn.csr
+
+REQDN=reqDN.justEmail
+echo blah@example.com | ${REQ} -new -key ${srcdir}/server.key -out justmail.csr
+
+### don't put ${REQ} invocations after here
+
+for f in server client twocn caseless cnfirst missingcn justmail; do
   ${CA} -days 900 -in ${f}.csr -out ${f}.cert
 done
 
-${CA} -extensions altExt -days 900 -in altname.csr -out altname.cert
-${CA} -extensions altExt2 -days 900 -in altname2.csr -out altname2.cert
-${CA} -extensions altExt3 -days 900 -in altname3.csr -out altname3.cert
-${CA} -extensions altExt4 -days 900 -in altname4.csr -out altname4.cert
+for n in 1 2 3 4 5; do
+ ${CA} -extensions altExt${n} -days 900 \
+     -in altname${n}.csr -out altname${n}.cert
+done
+
+MKPKCS12="${OPENSSL} pkcs12 -export -passout stdin -in client.cert -inkey client.key"
 
 # generate a PKCS12 cert from the client cert: -passOUT because it's the
 # passphrase on the OUTPUT cert, confusing...
-echo foobar | ${OPENSSL} pkcs12 -export -passout stdin -name "Neon Client Cert" \
-   -in client.cert -inkey client.key -out client.p12
+echo foobar | ${MKPKCS12} -name "Just A Neon Client Cert" -out client.p12
+
+# generate a PKCS#12 cert with no password and a friendly name
+echo | ${MKPKCS12} -name "An Unencrypted Neon Client Cert" -out unclient.p12
+
+# generate a PKCS#12 cert with no friendly name
+echo | ${MKPKCS12} -out noclient.p12
+
+### a file containing a complete chain
+
+cat ca/cert.pem server.cert > chain.pem
