@@ -1,6 +1,6 @@
 /* 
    HTTP request handling tests
-   Copyright (C) 2001-2002, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2001-2003, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -129,7 +129,7 @@ static int expect_header_value(const char *name, const char *value,
     
     ne_request_destroy(req);
     ne_session_destroy(sess);
-    free(gotval);
+    ne_free(gotval);
 
     return OK;
 }
@@ -156,7 +156,7 @@ static int expect_response(const char *expect, server_fn fn, void *userdata)
     return OK;
 }
 
-#define FOLLOW_REQ RESP200 "Content-Length: 0\r\n\r\n"
+#define EMPTY_RESP RESP200 "Content-Length: 0\r\n\r\n"
 
 /* Process a request with given method and response, expecting to get
  * a zero-length response body.  A second request is sent down the
@@ -166,14 +166,14 @@ static int expect_response(const char *expect, server_fn fn, void *userdata)
 static int expect_no_body(const char *method, const char *resp)
 {
     ne_session *sess = ne_session_create("http", "localhost", 7777);
-    ne_request *req = ne_request_create(sess, method, "/foobar");
+    ne_request *req = ne_request_create(sess, method, "/first");
     ssize_t ret;
-    char *r = ne_malloc(strlen(resp) + sizeof(FOLLOW_REQ));
+    char *r = ne_malloc(strlen(resp) + sizeof(EMPTY_RESP));
     
     strcpy(r, resp);
-    strcat(r, FOLLOW_REQ);
+    strcat(r, EMPTY_RESP);
     ON(spawn_server(7777, single_serve_string, r));
-    free(r);
+    ne_free(r);
 
     ONN("failed to begin request", ne_begin_request(req));
     ret = ne_read_response_block(req, buffer, BUFSIZ);
@@ -182,7 +182,7 @@ static int expect_no_body(const char *method, const char *resp)
 
     /* process following request; makes sure that nothing extra has
      * been eaten by the first request. */
-    ONV(any_request(sess, "/foo"),
+    ONV(any_request(sess, "/second"),
 	("second request on connection failed: %s",ne_get_error(sess)));
 
     ON(await_server());
@@ -257,6 +257,18 @@ static int no_body_HEAD(void)
 			  "Content-Length: 5\r\n\r\n");
 }
 
+static int no_body_empty_clength(void)
+{
+    return expect_no_body("GET", "HTTP/1.1 200 OK\r\n"
+			  "Content-Length:\r\n\r\n");
+}
+
+static int no_body_bad_clength(void)
+{
+    return expect_no_body("GET", "HTTP/1.1 200 OK\r\n"
+			  "Content-Length: foobar\r\n\r\n");
+}
+
 static int no_headers(void)
 {
     return expect_response("abcde", single_serve_string,
@@ -286,6 +298,15 @@ static int te_header(void)
 			   "\r\n" ABCDE_CHUNKS);
 }
 
+/* test that the presence of *any* t-e header implies a chunked
+ * response. */
+static int any_te_header(void)
+{
+    return expect_response("abcde", single_serve_string, RESP200
+                           "Transfer-Encoding: punked\r\n" "\r\n"
+                           ABCDE_CHUNKS);
+}
+
 static int chunk_numeric(void)
 {    
     /* leading zero's */
@@ -342,7 +363,7 @@ static int chunk_oversize(void)
     CALL(expect_response(body, single_serve_string, buf->data));
     
     ne_buffer_destroy(buf);
-    free(body);
+    ne_free(body);
 
     return OK;
 }
@@ -357,6 +378,17 @@ static int te_over_clength(void)
 			   ABCDE_CHUNKS);
 }
 
+/* te_over_clength with the headers the other way round; check for
+ * ordering problems. */
+static int te_over_clength2(void)
+{   
+    return expect_response("abcde", single_serve_string,
+			   RESP200 "Content-Length: 300\r\n" 
+			   TE_CHUNKED
+			   "\r\n"
+			   ABCDE_CHUNKS);
+}
+
 /* obscure case which is possibly a valid request by 2616, but should
  * be handled correctly in any case.  neon <0.22.0 tries to 
  * eat the response body, which is probably incorrect. */
@@ -475,6 +507,16 @@ static int fail_eof_headers(void)
     return fail_early_eof("HTTP/1.1 200 OK\r\nJimbob\r\n");
 }
 
+static int fail_eof_chunk(void)
+{
+    return fail_early_eof(RESP200 TE_CHUNKED "\r\n" "1\r\n" "a");
+}
+
+static int fail_eof_badclen(void)
+{
+    return fail_early_eof(RESP200 "Content-Length: 10\r\n\r\n" "abcde");
+}
+
 /* Persistent connection timeout where a FIN is sent to terminate the
  * connection, which is caught by a 0 return from the read() when the
  * second request reads the status-line. */
@@ -852,7 +894,7 @@ static int want_body(ne_socket *sock, void *userdata)
 
     ON(memcmp(buf, b->body, b->size));
     
-    free(buf);
+    ne_free(buf);
     return OK;
 }
 
@@ -919,7 +961,7 @@ static int send_bodies(void)
 	}
     }
 
-    free(bodies[BIG].body);
+    ne_free(bodies[BIG].body);
     return OK;
 }
 
@@ -937,8 +979,10 @@ static int serve_infinite_headers(ne_socket *sock, void *userdata)
 }
 
 /* Utility function: run a request using the given server fn, and the
- * request should fail. */
-static int fail_request(int with_body, server_fn fn, void *ud, int forever)
+ * request should fail. If 'error' is non-NULL, it must be a substring
+ * of the error string. */
+static int fail_request_with_error(int with_body, server_fn fn, void *ud, 
+                                   int forever, const char *error)
 {
     ne_session *sess = ne_session_create("http", "localhost", 7777);
     ne_request *req;
@@ -969,13 +1013,37 @@ static int fail_request(int with_body, server_fn fn, void *ud, int forever)
 	/* reap the server, don't care what it's doing. */
 	reap_server();
     }
+
+    NE_DEBUG(NE_DBG_HTTP, "Response gave error `%s'\n", ne_get_error(sess));
  
+    ONV(error && strstr(ne_get_error(sess), error) == NULL,
+        ("failed with error `%s', no `%s'", ne_get_error(sess), error));
+
+    if (!forever)
+        ONV(any_request(sess, "/fail/to/connect") != NE_CONNECT,
+            ("subsequent request re-used connection?"));
+
     ne_request_destroy(req);
     ne_session_destroy(sess);
    
     return OK;    
 }
 
+/* Run a random GET request which is given 'body' as the response; the
+ * request must fail, and 'error' must be found in the error
+ * string. */
+static int invalid_response_gives_error(const char *resp, const char *error)
+{
+    return fail_request_with_error(0, single_serve_string, (void *)resp, 0, error);
+}
+
+/* Utility function: run a request using the given server fn, and the
+ * request must fail. */
+static int fail_request(int with_body, server_fn fn, void *ud, int forever)
+{
+    return fail_request_with_error(with_body, fn, ud, forever, NULL);
+}
+
 static int unbounded_headers(void)
 {
     return fail_request(0, serve_infinite_headers, NULL, 0);
@@ -1024,12 +1092,13 @@ static int serve_close(ne_socket *sock, void *ud)
 static int is_alive(int port)
 {
     ne_sock_addr *addr;
-    ne_socket *sock = NULL;
-    ne_inet_addr *ia;
+    ne_socket *sock = ne_sock_create();
+    const ne_inet_addr *ia;
+    int connected = 0;
 
     addr = ne_addr_resolve("localhost", 0);
-    for (ia = ne_addr_first(addr); ia && !sock; ia = ne_addr_next(addr))
-	sock = ne_sock_connect(ia, 7777);
+    for (ia = ne_addr_first(addr); ia && !connected; ia = ne_addr_next(addr))
+	connected = ne_sock_connect(sock, ia, 7777) == 0;
     ne_addr_destroy(addr);
     if (sock == NULL)
 	return 0;
@@ -1216,6 +1285,19 @@ static int fail_lookup(void)
     return fail_noserver("no.such.domain", 7777, NE_LOOKUP);
 }
 
+/* neon 0.23.0 to 0.23.3: if a nameserver lookup failed, subsequent
+ * requests on the session would crash. */
+static int fail_double_lookup(void)
+{
+     ne_session *sess = ne_session_create("http", "nohost.example.com", 80);
+     ONN("request did not give lookup failure",
+	 any_request(sess, "/foo") != NE_LOOKUP);
+     ONN("second request did not give lookup failure",
+	 any_request(sess, "/bar") != NE_LOOKUP);
+     ne_session_destroy(sess);
+     return OK;
+}
+
 static int fail_connect(void)
 {
     return fail_noserver("localhost", 7777, NE_CONNECT);
@@ -1335,6 +1417,163 @@ static int fail_statusline(void)
     return OK;    
 }
 
+#define LEN (9000)
+static int fail_long_header(void)
+{
+    char resp[LEN + 500] = "HTTP/1.1 200 OK\r\n"
+	"Server: fish\r\n";
+    size_t len = strlen(resp);
+    
+    /* add a long header */
+    memset(resp + len, 'a', LEN);
+    resp[len + LEN] = '\0';
+    
+    strcat(resp, "\r\n\r\n");
+
+    return invalid_response_gives_error(resp, "Line too long");
+}
+
+static int fail_corrupt_chunks(void)
+{
+    static const struct {
+        const char *resp, *error;
+    } ts[] = {
+        /* not CRLF */
+        { RESP200 TE_CHUNKED "\r\n" "5\r\n" "abcdeFISH", 
+          "delimiter was invalid" },
+        /* short CRLF */
+        { RESP200 TE_CHUNKED "\r\n" "5\r\n" "abcde\n",
+          "not read chunk delimiter" },
+        /* CR-notLF */
+        { RESP200 TE_CHUNKED "\r\n" "5\r\n" "abcde\rZZZ",
+          "delimiter was invalid" },
+        { NULL, NULL }
+    };
+    int n;
+
+    for (n = 0; ts[n].resp; n++)
+        CALL(invalid_response_gives_error(ts[n].resp, ts[n].error));
+
+    return OK;
+}
+
+static int versions(void)
+{
+    ne_session *sess;
+
+    CALL(make_session(&sess, single_serve_string, 
+		      "HTTP/1.1 200 OK\r\n"
+		      "Content-Length: 0\r\n\r\n"
+
+		      "HTTP/1.0 200 OK\r\n"
+		      "Content-Length: 0\r\n\r\n"));
+    
+    CALL(any_request(sess, "/http11"));
+	 
+    ONN("did not detect HTTP/1.1 compliance",
+	ne_version_pre_http11(sess) != 0);
+	 
+    CALL(any_request(sess, "/http10"));
+
+    ONN("did not detect lack of HTTP/1.1 compliance",
+	ne_version_pre_http11(sess) == 0);
+
+    ne_session_destroy(sess);
+
+    return OK;
+}
+
+struct cr_args {
+    const char *method, *uri;
+    int result;
+};
+
+static void hk_createreq(ne_request *req, void *userdata,
+			 const char *method, const char *requri)
+{
+    struct cr_args *args = userdata;
+    
+    args->result = 1; /* presume failure */
+    
+    if (strcmp(args->method, method))
+	t_context("Hook got method %s not %s", method, args->method);
+    else if (strcmp(args->uri, requri))
+	t_context("Hook got Req-URI %s not %s", requri, args->uri);
+    else
+	args->result = 0;
+}
+
+static int hook_create_req(void)
+{
+    ne_session *sess;
+    struct cr_args args;
+
+    CALL(make_session(&sess, single_serve_string, EMPTY_RESP EMPTY_RESP));
+
+    ne_hook_create_request(sess, hk_createreq, &args);
+
+    args.method = "GET";
+    args.uri = "/foo";
+    args.result = -1;
+
+    CALL(any_request(sess, "/foo"));
+    
+    ONN("first hook never called", args.result == -1);
+    if (args.result) return FAIL;
+
+    args.uri = "http://localhost:7777/bar";
+    args.result = -1;
+    
+    /* force use of absoluteURI in request-uri */
+    ne_session_proxy(sess, "localhost", 7777);
+
+    CALL(any_request(sess, "/bar"));
+    
+    ONN("second hook never called", args.result == -1);
+    if (args.result) return FAIL;
+
+    ne_session_destroy(sess);
+
+    return OK;    
+}
+
+static int serve_check_method(ne_socket *sock, void *ud)
+{
+    char *method = ud;
+    char buf[20];
+    size_t methlen = strlen(method);
+
+    if (ne_sock_read(sock, buf, methlen) != (ssize_t)methlen)
+        return -1;
+    
+    ONN("method corrupted", memcmp(buf, method, methlen));
+    
+    return single_serve_string(sock, "HTTP/1.1 204 OK\r\n\r\n");
+}
+                             
+
+/* Test that the method string passed to ne_request_create is
+ * strdup'ed. */
+static int dup_method(void)
+{
+    char method[] = "FOO";
+    ne_session *sess;
+    ne_request *req;
+
+    CALL(make_session(&sess, serve_check_method, method));
+    
+    req = ne_request_create(sess, method, "/bar");
+    
+    strcpy(method, "ZZZ");
+
+    ONREQ(ne_request_dispatch(req));
+    ne_request_destroy(req);
+    ne_session_destroy(sess);
+    CALL(await_server());
+
+    return OK;
+}
+
 ne_test tests[] = {
     T(lookup_localhost),
     T(single_get_clength),
@@ -1343,15 +1582,19 @@ ne_test tests[] = {
     T(no_body_204),
     T(no_body_304),
     T(no_body_HEAD),
+    T(no_body_empty_clength),
+    T(no_body_bad_clength),
     T(no_headers),
     T(chunks),
     T(te_header),
+    T(any_te_header),
     T(reason_phrase),
     T(chunk_numeric),
     T(chunk_extensions),
     T(chunk_trailers),
     T(chunk_oversize),
     T(te_over_clength),
+    T(te_over_clength2),
     T(no_body_chunks),
     T(persist_http11),
     T(persist_chunked),
@@ -1386,13 +1629,21 @@ ne_test tests[] = {
     T(not_http),
     T(fail_eof_continued),
     T(fail_eof_headers),
+    T(fail_eof_chunk),
+    T(fail_eof_badclen),
+    T(fail_long_header),
+    T(fail_corrupt_chunks),
     T(read_timeout),
     T(fail_lookup),
+    T(fail_double_lookup),
     T(fail_connect),
     T(proxy_no_resolve),
     T(fail_chunksize),
     T(abort_respbody),
     T(retry_after_abort),
     T(fail_statusline),
+    T(dup_method),
+    T(versions),
+    T(hook_create_req),
     T(NULL)
 };
