@@ -1,6 +1,6 @@
 /* 
    Utility functions for HTTP client tests
-   Copyright (C) 2001-2002, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2001-2003, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -98,7 +98,7 @@ int any_2xx_request_body(ne_session *sess, const char *uri)
     int ret;
     ne_set_request_body_buffer(req, body, BSIZE);
     ret = ne_request_dispatch(req);
-    free(body);
+    ne_free(body);
     ONV(ret != NE_OK || ne_get_status(req)->klass != 2,
 	("request failed: %s\n", ne_get_error(sess)));
     ne_request_destroy(req);
